/* 
   Anagram program by Brian Scearce (c) 1990, 1991
   This program may be freely distributed provided this
   header remains intact.
*/

/* made the trivial changes so that you invoke it in the
 * expected style, namely:
 *     <program> string-to-anagram [min-len] 
 *
 * I don't know why Brian didn't write it that way in the first place!
 * Peter van der Linden, 1992
 */

typedef unsigned int at_least_26_bits; /* Change to long if you have to */

/*
  How it works: a dictionary is loaded in, and each of the words
  is put into a "letters" structure.  This structure, in addition
  to keeping the word itself, makes a bitmask of the letters in
  the word, an array showing how many of each letter is in the
  word, and a total of the number of letters in the word.

  When the user types in a phrase, it is made into a structure of
  the same kind.  So if I type in "Brian Scearce", the mask shows
  that the phrase contains "abceinrs" (in the bit mask), made up
  of 2 a's, 1 b, etc.  and that it contains 12 letters in all.
  The original word is saved, but the case of the letters and any
  non-letters do not matter.

  The dictionary is "stripped" -- that is, a new dictionary is
  created, and any words from the main dictionary that cannot
  possibly be in an anagram are not included.  For example, a word
  with an "x" in it cannot be in *any* anagram of "Brian Scearce",
  so it is not included in the stripped dictionary.  Nor can the
  word "babble", since it contains three b's.

  Now, the stripped dictionary is used to find anagrams.  A recursive
  routine (FindAnagram) is called that checks to see if the current
  word has a subset of letters from the anagramee.  For example,
  assume that the first word in the stripped dictionary is "barn".
  It is seen that this word is a subset of "Brian Scearce", so
  "barn" is added to the word list, and FindAnagram is called again
  with (effectively) "iscearce" as the new word.  When FindAnagram
  finds a *perfect* anagram, it prints out the entire word list.
  When it hits the end of the stripped dictionary, it returns.

  There are some other clever features put in to attempt to minimise
  the searching time in the stripped dictionary.  The *first* call
  (i.e. the top level of the recursion tree) attempts to match the
  most difficult letter.  The most difficult letter is the least
  common one in the stripped dictionary.  In the case of "Brian
  Scearce", it is probably "b".  This decreases the width of the
  search tree, which (probably?) speeds up the searching.  The
  least common letter is printed out as the "root letter" once the
  stripped dictionary is created.  The stripped dictionary is sorted
  so that all the words containing the root letter come first.

  Assume the stripped dictionary has N entries, the first K of them
  containing the root letter.

  If the root letter occurs exactly once, then the first level of
  the search tree need only look in the first K entries of the
  dictionary, and the rest of the search need only look in the last
  N-K entries.  This speeds things up.  I hope.

  The number that you type in on the command line (or by itself
  when the program is running) is the length of the shortest word
  you want in the stripped dictionary.  This ensures that you don't
  get boring anagrams consisting of short words.

  When the stripped dictionary is formed, it makes sure that all
  words included are longer than this shortest length, and not
  bigger than half of the shortest length (since you'd need a
  too-short word to fill in the rest (unless the word is exactly
  an anagram if the phrase typed in)).

  This program eats CPU cycles like mad.  Be careful using it.  I
  got some angry letters from a sysadmin trying to anagramize "Brian
  Leslie Scearce" with minlength == 3 :-)

  The dictionary that you read in at the beginning should be a list
  of newline seperated words.  Words are .*[a-z].*  They should
  have at least one letter.  The length of a word is the number of
  letters in it, so "he'll" has length 4.  I recommend you use
  /usr/dict/words if you have it, or do what I did: make your own
  file, starting with /usr/dict/words and adding stuff from normal
  text (letters, usenet, etc).
*/
  
#include "stdio.h"
#include "stdlib.h"
#include "string.h"
#include "memory.h"
#include "ctype.h"

#define FALSE 0
#define TRUE 1
#define MAXWORDS 50000
#define MAXWORDLEN 30

struct letters {
  at_least_26_bits mask;
  char count[26];
  int total;
  char *word;
};

static struct letters Dictionary[MAXWORDS];
static struct letters *StrippedDictionary[MAXWORDS];
int DictionaryEntries, SDEntries;
int MinimumLength = 3;

void DoAnagrams(char *av);
void Structify();
void DoAnagrams();
void RootFindAnagram();
void CreateStrippedDictionary();
void PrintStrippedDictionary();
int SortStrippedDictionary();
void FindAnagram();
int LettersOK();
int ReadDict();

main(ac, av)
int ac;
char *av[];
{
  if (ac != 2 && ac != 3 && ac!=4)
    (void) fprintf(stderr, "Usage: %s string-to-anagram [ min-len [wordfile] ] \n", av[0]);
  else {
    if (ac >= 3) MinimumLength = atoi(av[2]);
    if (ReadDict(ac==4?av[3]:"/usr/dict/words") != 0) 
        (void) fprintf(stderr, "Cannot read /usr/dict/words \n");
    else DoAnagrams(av[1]);
  }
  return 0;
}


int ReadDict(s)
char *s;
{
  FILE *fp;
  unsigned j;
  char buffer[MAXWORDLEN];
  int ch;

  if ((fp = fopen(s, "r")) == NULL) return 1;
  else {
    (void) fprintf(stderr, "reading dictionary...\n");
    ch = 0;
    while (ch != EOF) {
      for (j = 0; (ch = fgetc(fp)) != EOF; j++)
        if (ch == '\n') {
          buffer[j] = 0; 
          break;
        } else buffer[j] = ch;

        if ( (Dictionary[DictionaryEntries].word = malloc(j + 1)) == NULL) {
          (void) fprintf(stderr, "Out of memory on word %d\n", DictionaryEntries);
          return 1;
        }
        Structify(&Dictionary[DictionaryEntries], buffer);
        DictionaryEntries++;
    }
  }
  (void) fprintf(stderr, "main dictionary has %d entries\n", DictionaryEntries);
  return 0;
}


void Structify(Entry, s)
struct letters *Entry;
char *s;
{
  int i;
  int ch;

  (void) strcpy(Entry->word, s);
  Entry->mask = (at_least_26_bits) ~0;
  Entry->total = 0;
  for (i = 'a'; i <= 'z'; i++) Entry->count[i-'a'] = 0;
  for (i = strlen(s)-1 ; i >= 0; i--) {
    if (isupper(s[i])) ch = tolower(s[i]);
    else ch = s[i];
    ch -= 'a';
    if (ch >= 0 && ch < 26) {
      Entry->total++;
      Entry->count[ch]++;
      Entry->mask &= ~(1 << ch);
    }
  }
}

void DoAnagrams(char *av)
{
  char buffer[255];
  struct letters MyEntry;
  char *Words[50];

  (void) strcpy(buffer, av);

    if (isdigit(*buffer)) {
      MinimumLength = atoi(buffer);
    /*  (void) printf("New length: %d\n", MinimumLength); */
    } else if (*buffer == '?') {
      PrintStrippedDictionary();
    } else {
      MyEntry.word = buffer;
      Structify(&MyEntry, buffer);
      MyEntry.mask = ~MyEntry.mask;
      RootFindAnagram(MyEntry, Words);
      (void) printf("--%s------\n", buffer);
    }
}

void RootFindAnagram(Entry, Words)
struct letters Entry;
char *Words[];
{
  int RootIndexEnd;
 
  CreateStrippedDictionary(Entry);
  RootIndexEnd = SortStrippedDictionary(Entry);
  FindAnagram(Entry, Words, 0, 0, RootIndexEnd);
}

void CreateStrippedDictionary(Entry)
struct letters Entry;
{
  int i;

  for (SDEntries = i = 0; i < DictionaryEntries; i++)
    if (((Dictionary[i].mask | Entry.mask) == (unsigned)~0) &&
        (Dictionary[i].total >= MinimumLength) && 
        (Dictionary[i].total + MinimumLength <= Entry.total ||
         Dictionary[i].total == Entry.total) &&
        (LettersOK(Entry.count, Dictionary[i].count)))
      StrippedDictionary[SDEntries++] = &Dictionary[i];
/*   (void) fprintf(stderr, "stripped dictionary has %d\n", SDEntries); */
/*      PrintStrippedDictionary(); */
}

int LettersOK(EntryCount, DictionaryCount)
char EntryCount[], DictionaryCount[];
{
  int i;

  for (i = 25; i >= 0; i--)
    if (DictionaryCount[i] > EntryCount[i]) return FALSE;
  return TRUE;
}

void PrintStrippedDictionary()
{
  int i;

  for (i = 0; i < SDEntries; i++)
    (void) printf("%15s%c", StrippedDictionary[i]->word, (i%4 == 3)?'\n':' ');
  (void) printf("\n");
}

int SortStrippedDictionary(Entry)
struct letters Entry;
{
  extern at_least_26_bits SortMask;
  extern int StrangeCompare();
  int MasterCount[26];
  int LeastCommonIndex, LeastCommonCount;
  int i, j;

  for (j = 25; j >= 0; j--) MasterCount[j] = 0;

  for (i = SDEntries-1; i >= 0; i--)
    for (j = 25; j >= 0; j--)
      MasterCount[j] += StrippedDictionary[i]->count[j];

  LeastCommonCount = MAXWORDS * 5;
  for (j = 25; j >= 0; j--)
    if (MasterCount[j] < LeastCommonCount && MasterCount[j] != 0 &&
        (Entry.mask & (1 << j)) != 0) {
      LeastCommonCount = MasterCount[j];
      LeastCommonIndex = j;
    }
  SortMask = (1 << LeastCommonIndex);
  qsort((char *)StrippedDictionary, SDEntries, 
        sizeof(struct letters *), StrangeCompare); 
  for (i = 0; i < SDEntries; i++)
    if ((SortMask & ~StrippedDictionary[i]->mask) == 0)
      break;

  /* (void) fprintf(stderr, "key letter is '%c'\n", LeastCommonIndex + 'a');
  (void) fprintf(stderr, "root breadth is %d words\n", i);  */
  return i;
}

at_least_26_bits SortMask;
int StrangeCompare(s, t)
struct letters **s, **t;
{
  if ( ((*s)->mask & SortMask) != ((*t)->mask & SortMask))
    return (((*s)->mask & SortMask) > ((*t)->mask & SortMask))?1:-1;
  if ((*t)->total != (*s)->total)
    return ((*t)->total - (*s)->total);
  return strcmp((*s)->word, (*t)->word);
}



void FindAnagram(Entry, WordArray, Level, StartAt, EndAt)
struct letters Entry;
char **WordArray;
int Level;
int StartAt;
int EndAt;
{
  int i, j;
  struct letters EntryToPass;
  int Fail;

  for (i = StartAt; i < EndAt; i++) {
    if ( (Entry.mask | StrippedDictionary[i]->mask) != 0) {
      Fail = FALSE;
      for (j = 25; j >=0 && !Fail; j--)
        if (Entry.count[j] < StrippedDictionary[i]->count[j]) 
          Fail = TRUE;
      if (!Fail) {
        WordArray[Level] = StrippedDictionary[i]->word;
        EntryToPass.mask = 0;
        EntryToPass.total = 0;
        for (j = 25; j >= 0; j--) {
          if ( (EntryToPass.count[j] = 
                  Entry.count[j] - StrippedDictionary[i]->count[j]) != 0) {
            EntryToPass.total += EntryToPass.count[j];
            EntryToPass.mask |= 1 << j;
          }
        }
        if (EntryToPass.total == 0) { /* Found a series of words! */
          for (j = 0; j <= Level; j++) 
            (void) printf("%s ", WordArray[j]);
          (void) printf("\n");
        } else if (EntryToPass.total < MinimumLength) {
          ; /* Don't call again */
        } else {
          FindAnagram(EntryToPass, WordArray, Level+1, 
                         i, SDEntries);
        }
      }
    }
  }
}

